#include <linux/types.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <asm/io.h>
#include <linux/interrupt.h>
#include <linux/wrapper.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <asm/system.h>
#if LINUX_VERSION_CODE >= 0x020300
#include <linux/spinlock.h>
#else
#include <asm/spinlock.h>
#endif

#include "zr36060.h"/* There are some functions that need the ZR36060 status */
#include "zr36057.h"
#include "tvnorm.h"
/* gz: I am not really sure that all the overlay functions have been integrated
       and videointerlaced exists in this struct, but hasn't be checked for
       correct handling in all the functions */
/* I strongly recommend you to read the ZR36057/67 PDF manual from www.zoran.com,
   (especially page 21-23, about memory handling, if you use the "fake" ZR36057)
   otherwise you won't be able to understand the frag table organization */

#define GPIO_MASK 0xdf /* for the Buz code */

#define ZR57_NAME      "ZR36057/mgavideo"   /* name of the driver */

#define ZR57_ERR       KERN_ERR     ZR57_NAME
#define ZR57_DEBUG     KERN_INFO    ZR57_NAME
#define ZR57_INFO      KERN_INFO    ZR57_NAME
#define ZR57_WARNING   KERN_WARNING ZR57_NAME

#define MAX_KMALLOC_MEM (128*1024) /* the biggest memory chunk the kernel can allocate at once */

void zr36057_fake_mjpeg_enable_callback(void *cardstruct, int mode)
{
  printk(ZR57_ERR "This callback for mjpeg_enable should NEVER be called ! (mode = 0x%x)\n", mode);
  printk(ZR57_ERR "(There must be some internal error in the driver)\n");
}

void zr36057_fake_mjpeg_go_callback(void *cardstruct, int mode)
{
  printk(ZR57_ERR "This callback for mjpeg_go should NEVER be called ! (mode = 0x%x)\n", mode);
  printk(ZR57_ERR "(There must be some internal error in the driver)\n");
}

void zr36057_fake_or_callback(int value, int reg)
{
  printk(ZR57_ERR "This callback for OR should NEVER be called ! (val=0x%x, reg=0x%x)\n", 
	 value, reg);
  printk(ZR57_ERR "(There must be some internal error in the driver)\n");
}

void zr36057_fake_and_callback(int value, int reg)
{
  printk(ZR57_ERR "This callback for AND should NEVER be called ! (val=0x%x, reg=0x%x)\n", 
	 value, reg);
  printk(ZR57_ERR "(There must be some internal error in the driver)\n");
}

void zr36057_fake_write_callback(int value, int reg)
{
  printk(ZR57_ERR "This callback for WRITE should NEVER be called ! (val=0x%x, reg=0x%x)\n", 
	 value, reg);
  printk(ZR57_ERR "(There must be some internal error in the driver)\n");
}

int zr36057_init(struct zr36057_dev *zr36057, enum zr36057_mode real_chip, int triton, 
		 void *orfunction, void * andfunction, void * writefunction,
		 void *mjpeg_enable_function, void *mjpeg_go_function)
{
  unsigned long mem;
  unsigned mem_needed;
  int j;

  printk(KERN_INFO "init: Starting ZR36057 (%s mode)\n", 
	 (real_chip == REAL_ZR36057) ? "real" : "fake");

  zr36057->realchip = real_chip;
  zr36057->nbufs   = 0;
  zr36057->bufsize = 0;
  zr36057->buffers_allocated = 0;
  zr36057->stat_com = 0;
  zr36057->video_interlace = 0;
  zr36057->triton = triton;

  if (orfunction)
    zr36057->or = orfunction;
  else 
    zr36057->or = zr36057_fake_or_callback;

  if (andfunction)
    zr36057->and = andfunction;
  else 
    zr36057->and = zr36057_fake_and_callback;

  if (writefunction)
    zr36057->write = writefunction;
  else 
    zr36057->write = zr36057_fake_write_callback;

  if (mjpeg_enable_function)
    zr36057->mjpeg_enable = mjpeg_enable_function;
  else 
    zr36057->mjpeg_enable = zr36057_fake_mjpeg_enable_callback;

  if (mjpeg_go_function)
    zr36057->mjpeg_go = mjpeg_go_function;
  else 
    zr36057->mjpeg_go = zr36057_fake_mjpeg_go_callback;

  /* allocate memory *before* doing anything to the hardware in case allocation fails */  
  /* STAT_COM table and overlay mask */
  
  mem_needed = (MJPEG_NUM_STAT_COM + ((OVL_MAX_WIDTH + 31) / 32) * OVL_MAX_HEIGHT)*4;
  mem = (unsigned long) kmalloc(mem_needed, GFP_KERNEL);
  if (!mem) {
    printk(ZR57_ERR "zr36057_init: kmalloc (STAT_COM + ovl.mask) failed\n");
    return -ENOMEM;
  }
  memset((void*)mem, 0, mem_needed);
  
  zr36057->stat_com = (u32*)mem;
  for (j = 0; j < MJPEG_NUM_STAT_COM; j++) 
    zr36057->stat_com[j] = 1;      /* mark as unavailable to zr36057 */
  
  /* Assign the overlay mask its new memory */
  zr36057->overlay_mask = (u32 *)(mem + MJPEG_NUM_STAT_COM*4);
  
  /* Initialize btv->jpg_gbuf */
  
  for (j = 0; j < MJPEG_MAX_FRAME; j++) 
    {
      zr36057->gbuf[j].frag_tab       = 0;
      zr36057->gbuf[j].frag_tab_bus   = 0;
      zr36057->gbuf[j].state          = MJPEG_STATE_USER;
      zr36057->gbuf[j].bs.frame       = j;
    }

   if (zr36057->realchip)
     { 
       /* reset zr36057 */
       zr36057->write(0, ZR36057_SPGPPCR);
       mdelay(10);

       /* take zr36057 out of reset now */
       zr36057->write(ZR36057_SPGPPCR_SoftReset, ZR36057_SPGPPCR);
       mdelay(10);

       /* stop all DMA processes */
       zr36057->write(ZR36057_MCTCR_CFlush, ZR36057_MCTCR);
       zr36057->and(~ZR36057_VDCR_VidEn, ZR36057_VDCR);
       /* assert P_Reset */
       zr36057->write(0, ZR36057_JPC);

       /* set up GPIO direction */
       zr36057->write(ZR36057_SPGPPCR_SoftReset | 0, ZR36057_SPGPPCR);

       /* Set up guest bus timing - Guests 0..3 Tdur=12, Trec=3 */
       zr36057->write((GPIO_MASK << 24) | 0x8888, ZR36057_GPPGCR1);
       mdelay(10);
     }

   return 0;
}

/* Sync on an MJPEG buffer (was: jpg_sync in buz.c) */
int zr36057_sync(struct zr36057_dev *zr36057, struct zr36060_dev *zr36060, 
		 struct zr36057_sync *bs, spinlock_t *lock)
{
   unsigned long flags;
   int frame;

   if( zr36060->state != ZR_DECOMPRESS_STATE &&
       zr36060->state != ZR_COMPRESS_STATE ) {
      printk(ZR57_ERR ": mgavideo_zr36060_sync: - codec not in streaming mode\n");
      return -EINVAL;
   }

   /* If we already have reached the end of the queue, wait until the transfer is done */ 
   while(zr36057->que_tail == zr36057->dma_tail) 
     {
       interruptible_sleep_on(&zr36057->capq);
       if (signal_pending(current)) return -ERESTARTSYS;
     }

   if (zr36057->realchip) /* only for real chips, fake ones just run in software */
     spin_lock_irqsave(lock, flags);

   frame = zr36057->pend[(zr36057->que_tail)++ & MJPEG_MASK_FRAME];

   /* buffer should now be in MJPEG_STATE_DONE */
   if(zr36057->gbuf[frame].state != MJPEG_STATE_DONE)
      printk(ZR57_ERR ": mgavideo_zr36060_sync - internal error\n");

   *bs = zr36057->gbuf[frame].bs;
   zr36057->gbuf[frame].state = MJPEG_STATE_USER;

   if (zr36057->realchip) /* only for real chips, fake ones just run in software */
     spin_unlock_irqrestore(lock, flags);

   return 0;
}

/* Queue a buffer for capture or playback. The first call also starts streaming capture. 
   When streaming capture is going on, you may only queue further buffers or 
   issue syncs until streaming capture is switched off again with a
   argument of -1 to a BUZIOC_QBUF_CAPT/BUZIOC_QBUF_PLAY ioctl. */
int zr36057_queuebuf_ops(struct zr36057_dev *zr36057, struct zr36060_dev *zr36060, 
			 u32 frame, int mode, void *hostcard)
{
  unsigned long flags;
  int res;
  
  /* Check if buffers are allocated */
  if(!zr36057->buffers_allocated)
    {
      printk(ZR57_ERR ": queuebuf_ops: buffers not yet allocated\n");
      return -ENOMEM;
    }
  
  /* Does the user want to stop streaming? */
  if (frame < 0) 
    {
      if (zr36060->state == mode) 
	{
	  (*zr36057->mjpeg_enable)(hostcard, ZR_IDLE_STATE);
	  return 0;
	}
      else
	{
	  printk(ZR57_ERR ": queuebuf_ops: command ""stop streaming"" but not in streaming mode\n");
	  return -EINVAL;
	}
    }

  /* No grabbing outside the buffer range! */
  if (frame >= zr36057->nbufs) 
    {
      printk(ZR57_ERR ": queuebuf_ops: buffer %d out of range\n",frame);
      return -EINVAL;
    }
  
  /* what is the codec mode right now? */
  if ( zr36060->state == ZR_IDLE_STATE )
    {
      /* Ah, no grabbing was done yet - Ok load up the zr36060 and go */
      (*zr36057->mjpeg_enable)(hostcard, mode);
    }
  else if ( zr36060->state != mode )
    {
      /* wrong codec mode active - invalid */
      printk(ZR57_ERR ": queuebuf_ops - codec in wrong mode\n");
      return -EINVAL;
    }

  /* we are already grabbing (mode is equal) */
#ifdef NEVER
  if (zr36057->realchip) /* only for real chips, fake ones just run in software */
    spin_lock_irqsave(&mga->lock, flags);
#endif

  /* make sure a grab isn't going on currently with this buffer */
  switch (zr36057->gbuf[frame].state) 
    {
    default:
    case MJPEG_STATE_DMA:
    case MJPEG_STATE_PEND:
    case MJPEG_STATE_DONE:
      res = -EBUSY;         /* what are you doing? */
      break;

    case MJPEG_STATE_USER:
      /* since there is at least one unused buffer there's room for at least one more pend[] entry */
      zr36057->pend[(zr36057->que_head)++ & MJPEG_MASK_FRAME] = frame;
      zr36057->gbuf[frame].state = MJPEG_STATE_PEND;
      zr36057_feed_stat_com(zr36057, zr36060);
      res = 0;
      break;
    }

  if (zr36057->realchip) /* only for real chips, fake ones just run in software */
    spin_unlock_irqrestore(&mga->lock, flags);
  
  /* Start the zr36060 when the first frame is queued  */
  if(zr36057->que_head == 1)
    {
      printk("mgacap: NOW STARTING CODEC FOR TRANSFER\n");
      /* Make the MGA CODEC start the JPEG content transfer */  
      if (zr36057->realchip)
	{
	  zr36057->or(ZR36057_JMC_Go_en, ZR36057_JMC);
	  zr36057->write(ZR36057_JPC_P_Reset | ZR36057_JPC_CodTrnsEn 
			 | ZR36057_JPC_Active, ZR36057_JPC);
	}
      else
	(*zr36057->mjpeg_go)(hostcard, mode);
    }

  return res;
}

/*
 *   Allocate the MJPEG grab buffers.
 *
 *   If the requested buffer size is smaller than MAX_KMALLOC_MEM,
 *   kmalloc is used to request a physically contiguous area,
 *   else we allocate the memory in framgents with get_free_page.
 *
 *   If a Natoma chipset is present and this is a revision 1 zr36057,
 *   each MJPEG buffer needs to be physically contiguous.
 *   (RJ: This statement is from Dave Perks' original driver,
 *   I could never check it because I have a zr36067)
 *   The driver cares about this because it reduces the buffer
 *   size to MAX_KMALLOC_MEM in that case (which forces contiguous allocation).
 *
 *   RJ: The contents grab buffers needs never be accessed in the driver.
 *       Therefore there is no need to allocate them with vmalloc in order
 *       to get a contiguous virtual memory space.
 *       I don't understand why many other drivers first allocate them with
 *       vmalloc (which uses internally also get_free_page, but delivers you
 *       virtual addresses) and then again have to make a lot of efforts
 *       to get the physical address.
 *
 */

 int zr36057_fbuffer_alloc(struct zr36057_dev *zr36057)
{
   int i, j, off, alloc_contig;
   unsigned long mem;

   /* Decide if we should alloc contiguous or fragmented memory */
   /* This has to be identical in jpg_fbuffer_alloc and jpg_fbuffer_free */

   alloc_contig = (zr36057->bufsize < MAX_KMALLOC_MEM);

   for (i = 0; i < zr36057->nbufs; i++)
   {
      if(zr36057->gbuf[i].frag_tab)
         printk(ZR57_WARNING ": zr36057_fbuffer_alloc: buffer %d already allocated ???\n",i);

      /* Allocate fragment table for this buffer */

      mem = get_free_page(GFP_KERNEL);
      if (mem==0) 
      {
         printk(ZR57_ERR ": zr36057_fbuffer_alloc: get_free_page (frag_tab) failed for buffer %d\n",i);
         zr36057_fbuffer_free(zr36057);
         return -ENOBUFS;
      }
      memset((void*)mem,0,PAGE_SIZE);
      zr36057->gbuf[i].frag_tab     = (u32*) mem;
      zr36057->gbuf[i].frag_tab_bus = virt_to_bus((void*)mem);

      if (alloc_contig)
      {
	mem = (unsigned long)kmalloc(zr36057->bufsize,GFP_KERNEL); /* Get one big chunk of memory */
         if(mem==0)
         {
            printk(ZR57_ERR ": zr36057_fbuffer_alloc: kmalloc failed for buffer %d\n",i);
            zr36057_fbuffer_free(zr36057);
            return -ENOBUFS;
         }
         zr36057->gbuf[i].frag_tab[0] = virt_to_bus((void*)mem); /* write the physical fragment address */
	 /* write buffersize ( | 1 means final fragment) */
         zr36057->gbuf[i].frag_tab[1] = ((zr36057->bufsize/4)<<1)|1; 
	 /* reserve the according memory pages */
         for(off=0;off<zr36057->bufsize;off+=PAGE_SIZE) mem_map_reserve(MAP_NR(mem+off));
      }
      else
      {
         /* jpg_bufsize is already page aligned */
         for (j=0; j<zr36057->bufsize/PAGE_SIZE; j++)
         {
            mem = get_free_page(GFP_KERNEL);
            if(mem==0)
            {
               printk(ZR57_ERR ": zr36057_fbuffer_alloc: get_free_page failed for buffer %d\n",i);
               zr36057_fbuffer_free(zr36057);
               return -ENOBUFS;
            }

            zr36057->gbuf[i].frag_tab[2*j  ] = virt_to_bus((void*)mem);
            zr36057->gbuf[i].frag_tab[2*j+1] = (PAGE_SIZE/4)<<1;
            mem_map_reserve(MAP_NR(mem));
         }

         zr36057->gbuf[i].frag_tab[2*j-1] |= 1;
      }
   }

   printk(ZR57_ERR "zr36057_fbuffer_alloc: %ld KB allocated\n",
	  (zr36057->nbufs * zr36057->bufsize)>>10);
   zr36057->buffers_allocated = 1;
   return 0;
}

/* free the MJPEG grab buffers */
void zr36057_fbuffer_free(struct zr36057_dev * zr36057)
{
   int i, j, off, alloc_contig;
   unsigned char *mem;

   /* Decide if we should free contiguous or fragmented memory */
   /* This has to be identical in jpg_fbuffer_alloc and jpg_fbuffer_free */
   alloc_contig = (zr36057->bufsize < MAX_KMALLOC_MEM);

   for (i = 0; i < zr36057->nbufs; i++) /* count through all MJPEG buffers */
   {
     if(!zr36057->gbuf[i].frag_tab) continue;
     
     if(alloc_contig)
       {
	 if(zr36057->gbuf[i].frag_tab[0]) /* there is just one fragment pair */
	   {
	     mem = (unsigned char *) bus_to_virt(zr36057->gbuf[i].frag_tab[0]); /* create app addresses */
	     for(off=0;off<zr36057->bufsize;off+=PAGE_SIZE) 
	       mem_map_unreserve(MAP_NR(mem+off));
	     kfree((void *) mem);
	     zr36057->gbuf[i].frag_tab[0] = 0; /* this is the physical address for the fragment */
	     zr36057->gbuf[i].frag_tab[1] = 0; /* and this the length */
	   }
       }
      else
	{
	  for(j=0; j<zr36057->bufsize/PAGE_SIZE; j++) /* each fragment if of 4k size */ 
	    {
	      if(!zr36057->gbuf[i].frag_tab[2*j]) break; /* until a fragment does not exist */
	      mem_map_unreserve(MAP_NR(bus_to_virt(zr36057->gbuf[i].frag_tab[2*j])));
	      free_page((unsigned long)bus_to_virt(zr36057->gbuf[i].frag_tab[2*j]));
	      zr36057->gbuf[i].frag_tab[2*j  ] = 0; 
	      zr36057->gbuf[i].frag_tab[2*j+1] = 0;
	    }
	}
     
     free_page( (unsigned long)zr36057->gbuf[i].frag_tab );
     zr36057->gbuf[i].frag_tab = NULL;
   }
   zr36057->buffers_allocated = 0;
}

/*
 * set geometry
 */
void zr36057_set_vfe(struct zr36057_dev *zr36057,
		     struct zr36060_dev *zr36060,
		     int video_width, int video_height, int windowwidth,
		     unsigned int video_format)
{
   struct tvnorm * tvn;
   unsigned HStart, HEnd, VStart, VEnd;
   unsigned DispMode;
   unsigned VidWinWid, VidWinHt;
   unsigned hcrop1, hcrop2, vcrop1, vcrop2;
   unsigned Wa, We, Ha, He;
   unsigned X, Y, HorDcm, VerDcm;
   u32 reg;
   unsigned mask_line_size;

   if (!zr36057->realchip) 
     { 
       printk(ZR57_WARNING "Video Front End settings are not possible in ZR36057 emulation !\n");
       return;
     }

   if(zr36060->params.norm<0 || zr36060->params.norm>1) {
      printk(ZR57_ERR ": set_vfe: video_norm = %d not valid\n", zr36060->params.norm);
      return;
   }
   if(video_width< OVL_MIN_WIDTH || video_height< OVL_MIN_HEIGHT) {
      printk(ZR57_ERR ": set_vfe: w=%d h=%d not valid\n", 
	     video_width, video_height);
      return;
   }

   tvn = &tvnorms[zr36060->params.norm];

   Wa = tvn->Wa;
   Ha = tvn->Ha;

   /* if window has more than half of active height,
      switch on interlacing - we want the full information */

   zr36057->video_interlace = (video_height > Ha / 2);

   /**** zr36057 ****/

   /* horizontal */
   VidWinWid = video_width;
   X = (VidWinWid * 64 + tvn->Wa - 1) / tvn->Wa;
   We = (VidWinWid * 64) / X;
   HorDcm = 64 - X;
   hcrop1 = 2 * ((tvn->Wa - We) / 4);
   hcrop2 = tvn->Wa - We - hcrop1;
   HStart = tvn->HStart | 1;
   HEnd = HStart + tvn->Wa - 1;
   HStart += hcrop1;
   HEnd -= hcrop2;
   reg = ((HStart & ZR36057_VFEHCR_Hmask) << ZR36057_VFEHCR_HStart)
       | ((HEnd & ZR36057_VFEHCR_Hmask)   << ZR36057_VFEHCR_HEnd);
   reg |= ZR36057_VFEHCR_HSPol;
   zr36057->write(reg, ZR36057_VFEHCR);

   /* Vertical */
   DispMode = !zr36057->video_interlace;
   VidWinHt = DispMode ? video_height : video_height / 2;
   Y = (VidWinHt * 64 * 2 + tvn->Ha - 1) / tvn->Ha;
   He = (VidWinHt * 64) / Y;
   VerDcm = 64 - Y;
   vcrop1 = (tvn->Ha / 2 - He) / 2;
   vcrop2 = tvn->Ha / 2 - He - vcrop1;
   VStart = tvn->VStart;
   VEnd = VStart + tvn->Ha / 2 - 1;
   VStart += vcrop1;
   VEnd -= vcrop2;
   reg = ((VStart & ZR36057_VFEVCR_Vmask) << ZR36057_VFEVCR_VStart)
       | ((VEnd & ZR36057_VFEVCR_Vmask)   << ZR36057_VFEVCR_VEnd);
   reg |= ZR36057_VFEVCR_VSPol;
   zr36057->write(reg, ZR36057_VFEVCR);

   /* scaler and pixel format */
   reg = 0 // ZR36057_VFESPFR_ExtFl /* Trying to live without ExtFl */
      | (HorDcm << ZR36057_VFESPFR_HorDcm)
      | (VerDcm << ZR36057_VFESPFR_VerDcm)
      | (DispMode << ZR36057_VFESPFR_DispMode)
      | ZR36057_VFESPFR_LittleEndian;
   /* RJ: I don't know, why the following has to be the opposite
          of the corresponding ZR36060 setting, but only this way
          we get the correct colors when uncompressing to the screen  */
   reg |= ZR36057_VFESPFR_VCLKPol;
   /* RJ: Don't know if that is needed for NTSC also */
   reg |= ZR36057_VFESPFR_TopField;
   switch (video_format) {

   case VIDEO_PALETTE_YUV422:
      reg |= ZR36057_VFESPFR_YUV422;
      break;

   case VIDEO_PALETTE_RGB555:
      reg |= ZR36057_VFESPFR_RGB555 | ZR36057_VFESPFR_ErrDif;
      break;

   case VIDEO_PALETTE_RGB565:
      reg |= ZR36057_VFESPFR_RGB565 | ZR36057_VFESPFR_ErrDif;
      break;

   case VIDEO_PALETTE_RGB24:
      reg |= ZR36057_VFESPFR_RGB888 | ZR36057_VFESPFR_Pack24;
      break;

   case VIDEO_PALETTE_RGB32:
      reg |= ZR36057_VFESPFR_RGB888;
      break;

   default:
      printk(ZR57_INFO ": Unknown color_fmt=%x\n", video_format);
      return;

   }
   if (HorDcm >= 48) {
      reg |= 3 << ZR36057_VFESPFR_HFilter; /* 5 tap filter */
   } else if (HorDcm >= 32) {
      reg |= 2 << ZR36057_VFESPFR_HFilter; /* 4 tap filter */
   } else if (HorDcm >= 16) {
      reg |= 1 << ZR36057_VFESPFR_HFilter; /* 3 tap filter */
   }
   zr36057->write(reg, ZR36057_VFESPFR);

   /* display configuration */

   reg = (16 << ZR36057_VDCR_MinPix)
       | (VidWinHt << ZR36057_VDCR_VidWinHt)
       | (VidWinWid << ZR36057_VDCR_VidWinWid);
   if (zr36057->triton)
      reg &= ~ZR36057_VDCR_Triton;
   else
      reg |= ZR36057_VDCR_Triton;
   zr36057->write(reg, ZR36057_VDCR);

   /* Write overlay clipping mask data, but don't enable overlay clipping */
   /* RJ: since this makes only sense on the screen, we use 
      zr36057->window.width instead of video_width */

   mask_line_size = (OVL_MAX_WIDTH + 31) / 32;
   reg = virt_to_bus(zr36057->overlay_mask);
   zr36057->write(reg, ZR36057_MMTR);
   reg = virt_to_bus(zr36057->overlay_mask + mask_line_size);
   zr36057->write(reg, ZR36057_MMBR);
   reg = mask_line_size - (windowwidth+31)/32;
   if (DispMode == 0) reg += mask_line_size;
   reg <<= ZR36057_OCR_MaskStride;
   zr36057->write(reg, ZR36057_OCR);

}

/*
 * zr36057_overlay was not integrated yet as there were
 * too many variables from the btv struct involved ... 
 */

/*
 * The overlay mask has one bit for each pixel on a scan line,
 *  and the maximum window size is OVL_MAX_WIDTH * OVL_MAX_HEIGHT pixels.
 */
void zr36057_write_overlay_mask(struct zr36057_dev * zr36057, 
				struct zr36057_video_clip * vp, int count,
				int windowwidth, int windowheight)
{
   unsigned mask_line_size = (OVL_MAX_WIDTH + 31) / 32;
   u32 * mask;
   int x, y, width, height;
   unsigned i, j, k;
   u32 reg;

   /* fill mask with "one bits" */
   memset(zr36057->overlay_mask, ~0, mask_line_size * 4 * OVL_MAX_HEIGHT);
   reg = 0;

   for (i = 0; i < count; ++i) {
      /* pick up local copy of clip */
      x = vp[i].x;
      y = vp[i].y;
      width  = vp[i].width;
      height = vp[i].height;

      /* trim clips that extend beyond the window */
      if (x < 0) {
         width += x;
         x = 0;
      }
      if (y < 0) {
         height += y;
         y = 0;
      }
      if (x + width > windowwidth) {
         width = windowwidth - x;
      }
      if (y + height > windowheight) {
         height = windowheight - y;
      }

      /* ignore degenerate clips */
      if (height <= 0) {
         continue;
      }
      if (width <= 0) {
         continue;
      }

      /* apply clip for each scan line */
      for (j = 0; j < height; ++j) {
         /* reset bit for each pixel */
         /* this can be optimized later if need be */
         mask = zr36057->overlay_mask + (y + j) * mask_line_size;
         for (k = 0; k < width; ++k) {
            mask[(x + k) / 32] &= ~((u32)1 << (x + k) % 32);
         }
      }
   }
}

void zr36057_set_jpg(struct zr36057_dev * zr36057, struct zr36060_dev *zr36060, int mode)
{
   struct tvnorm * tvn = NULL;
   u32 reg;
   int i;

   /* re-initialize DMA ring stuff */
   zr36057->que_head = 0;
   zr36057->dma_head = 0;
   zr36057->dma_tail = 0;
   zr36057->que_tail = 0;
   zr36057->seq_num  = 0;
   for (i = 0; i < MJPEG_NUM_STAT_COM; ++i) 
      zr36057->stat_com[i] = 1;      /* mark as unavailable to zr36057 */
   
   for (i = 0; i < zr36057->nbufs; i++) 
     zr36057->gbuf[i].state = MJPEG_STATE_USER;   /* nothing going on */

   if (zr36057->realchip)
     {
       tvn = &tvnorms[zr36060->params.norm];

       zr36057->write(0, ZR36057_JPC);   /* assert P_Reset */
       
       /* MJPEG compression mode */
       switch (mode) 
	 {
	 
	 case ZR_COMPRESS_STATE:
	 default:
	   reg = ZR36057_JMC_MJPGCmpMode;
	   break;
	   
	 case ZR_DECOMPRESS_STATE:
	   reg = ZR36057_JMC_MJPGExpMode;
	   reg |= ZR36057_JMC_SyncMstr;
	   /* RJ: The following is experimental - improves the output to screen */
	   if(zr36060->params.VFIFO_FB) reg |= ZR36057_JMC_VFIFO_FB;
	   break;
	   
	   /* These do not exist in the definition for the JPEG codec ! */
	   /*   case BUZ_MODE_STILL_COMPRESS: reg = ZR36057_JMC_JPGCmpMode; break; */
	   /*   case BUZ_MODE_STILL_DECOMPRESS: reg = ZR36057_JMC_JPGExpMode; break; */
	 }
       reg   |= ZR36057_JMC_JPG;
       if(zr36060->params.field_per_buff == 1) reg |= ZR36057_JMC_Fld_per_buff;
       zr36057->write(reg, ZR36057_JMC);
       
       /* vertical */
       zr36057->or(ZR36057_VFEVCR_VSPol, ZR36057_VFEVCR);
       reg = (6 << ZR36057_VSP_VsyncSize) | (tvn->Ht << ZR36057_VSP_FrmTot);
       zr36057->write(reg, ZR36057_VSP);
       reg = ( (zr36060->params.img_y+tvn->VStart) << ZR36057_FVAP_NAY)
	 | (zr36060->params.img_height << ZR36057_FVAP_PAY);
       zr36057->write(reg, ZR36057_FVAP);
       
       /* horizontal */
       zr36057->or(ZR36057_VFEHCR_HSPol,ZR36057_VFEHCR);
       reg = ( (tvn->Wt-100) << ZR36057_HSP_HsyncStart) | (tvn->Wt << ZR36057_HSP_LineTot);
       zr36057->write(reg, ZR36057_HSP);
       reg = ( (zr36060->params.img_x+tvn->HStart) << ZR36057_FHAP_NAX)
	 | (zr36060->params.img_width << ZR36057_FHAP_PAX);
       zr36057->write(reg, ZR36057_FHAP);
       
       /* field process parameters */
       if(zr36060->params.odd_even)
	 reg = ZR36057_FPP_Odd_Even;
       else
	 reg = 0;
       zr36057->write(reg, ZR36057_FPP);
       
       /* Set proper VCLK Polarity, else colors will be wrong during playback */
       zr36057->or(ZR36057_VFESPFR_VCLKPol, ZR36057_VFESPFR);
       
       /* code base address and FIFO threshold */
       reg = virt_to_bus(zr36057->stat_com);
       zr36057->write(reg, ZR36057_JCBA);
       reg = 0x50;
       zr36057->write(reg, ZR36057_JCFT);
       
       /* JPEG codec guest ID */
       reg = (1 << ZR36057_JCGI_JPEGuestID) | (0 << ZR36057_JCGI_JPEGuestReg);
       zr36057->write(reg, ZR36057_JCGI);
       
       /* Code transfer guest ID */
       reg = (0 << ZR36057_MCTCR_CodGuestID) | (3 << ZR36057_MCTCR_CodGuestReg);
       reg |= ZR36057_MCTCR_CFlush;
       zr36057->write(reg, ZR36057_MCTCR);
       
       /* deassert P_Reset */
       zr36057->write(ZR36057_JPC_P_Reset, ZR36057_JPC);
     }
   else
     {
       printk(ZR57_INFO "Resetting fake: stat_com_count, frag_num and frag_filled\n");
       zr36057->fake_stat_com_count = 0; /* reset the fake stat_com counter */ 
       zr36057->fake_frag_num = 0;
       zr36057->fake_frag_done = 0;
       /* MJPEG compression mode */
       switch (mode) 
	 {
	   
	 case ZR_COMPRESS_STATE:
	 default:
	   {
	     printk(ZR57_INFO "ZR36057 is prepared for compression\n");
	     zr36057->fake_seq_num = 0;
	     zr36057->fake_rec_bytes = 0;
	   };
	   break;
	   
	 case ZR_DECOMPRESS_STATE:
	   {
	     printk(ZR57_INFO "ZR36057 is prepared for decompression\n");
	   }
	   break;
	 }
     }
}

/* zr36057_enable_jpg (mjpeg_enable) must be done in the "host" driver (because the decoder must be programmed, too) */

/* Grant the user app access to the JPEG framebuffer */
int zr36057_mmap(struct zr36057_dev *zr36057, const char * adr, unsigned long size)
{
  unsigned long start = (unsigned long) adr;
  unsigned long page, pos, todo, fraglen;
  int i, j;
  
  /* Map the MJPEG buffers */
  if(!zr36057->buffers_allocated)
    {
      printk(ZR57_ERR ": mmap_mjpegbufs: buffers not yet allocated\n");
      return -ENOMEM;
    }
  
  if(size > zr36057->nbufs * zr36057->bufsize)
    {
      printk(ZR57_ERR ": mmap_mjpegbufs: Max size is %ld - you wanted %ld\n",
	     zr36057->nbufs * zr36057->bufsize, size);
      return -EINVAL;
    }
  
  if(size != zr36057->nbufs * zr36057->bufsize)
    printk(ZR57_WARNING ": mmap_mjpegbufs: Expected %ld - you wanted %ld\n",
	   zr36057->nbufs * zr36057->bufsize, size);
  
   for(i=0; i < zr36057->nbufs; i++) /* count through the user app buffers */
     {
       /* count through all pages that belong to a buffer */
       for(j=0 ; j < zr36057->bufsize / PAGE_SIZE ; j++) 
         {
	   /* Get a fragment of the buffer */
	   fraglen = (zr36057->gbuf[i].frag_tab[2*j+1] & ~1) << 1;
	   todo = size;	   /* try to take all the rest in one chunk */
	   if(todo > fraglen) /* simply cut if the chunk is too big for this time */
	     todo = fraglen;  /* it should fit evenly in the end (see warning message above) */
	   pos = (unsigned long)zr36057->gbuf[i].frag_tab[2*j];
	   page = virt_to_phys(bus_to_virt(pos)); /* should just be pos on i386 */
	   
	   if (remap_page_range(start, page, todo, PAGE_SHARED)) {
	     printk(ZR57_ERR ": mmap_mjpegbufs: remap_page_range failed\n");
	     return -EAGAIN;
	   }
	   size  -= todo; /* reduce it by the memory we just mapped */
	   start += todo; /* and increase the start address */
	   if(size==0) break; /* exit when complete */
	   if(zr36057->gbuf[i].frag_tab[2*j+1] & 1) break; /* was last fragment */
         }
       if(size==0) break; /* exit when complete */
     }
   return 0;
}

/* move frames from pending queue to DMA, or more explicitly: */
/* put a pending entry in the MJPEG user app buffer into the stat_com pointer, which 
   is DMA:ed, either by the real ZR36057 or by the fake ZR36057 handling routines */
/* when this is called the spinlock must be held */
void zr36057_feed_stat_com(struct zr36057_dev *zr36057, struct zr36060_dev *zr36060)
{
  int frame, i, max_stat_com;

  /* in full resolution, two consecutive buffers contain the even respectively odd field */
  max_stat_com = (zr36060->params.TmpDcm == 1) ? MJPEG_NUM_STAT_COM : (MJPEG_NUM_STAT_COM>>1);

  while ((zr36057->dma_head - zr36057->dma_tail) < max_stat_com /* while there is space for new buffers */
	 && zr36057->dma_head != zr36057->que_head) /* and we have not catched up yet */  
    {
      frame = zr36057->pend[zr36057->dma_head & MJPEG_MASK_FRAME]; /* get the latest pending entry */
      if(zr36060->params.TmpDcm == 1) 
      {
         /* fill 1 stat_com entry */
         i = zr36057->dma_head & MJPEG_MASK_STAT_COM;
         zr36057->stat_com[i] = zr36057->gbuf[frame].frag_tab_bus;
      }
      else /* odd/even fields (full res) */
      {
         /* fill 2 stat_com entries */
         i = (zr36057->dma_head & 1)*2;
         zr36057->stat_com[i  ] = zr36057->gbuf[frame].frag_tab_bus;
         zr36057->stat_com[i+1] = zr36057->gbuf[frame].frag_tab_bus;
      }
      zr36057->gbuf[frame].state = MJPEG_STATE_DMA;
      zr36057->dma_head++; /* advance the DMA head */
   }
}

/* move frames from DMA queue to done queue, or more explicitly: */
/* when this is called the spinlock must be held */
void zr36057_reap_stat_com(struct zr36057_dev *zr36057, struct zr36060_dev *zr36060)
{
   int i;
   u32 stat_com;
   unsigned int seq;
   unsigned int dif;
   int frame;
   struct zr36057_gbuffer * gbuf;

   /* In motion decompress we don't have a hardware frame counter,
      we just count the interrupts here */
   if (zr36060->state == ZR_DECOMPRESS_STATE) zr36057->seq_num++;

   while (zr36057->dma_tail != zr36057->dma_head) /* while we have not catched up */
   {
     if(zr36060->params.TmpDcm == 1) /* full res handling */
       i = zr36057->dma_tail & MJPEG_MASK_STAT_COM; /* "normal" operation */
     else
       i = (zr36057->dma_tail & 1)*2+1; /* for the odd/even pair buffers */

      stat_com = zr36057->stat_com[i];

      if ((stat_com & 1) == 0) /* & 1 == 0 means that this stat_com entry has not been serviced yet */
         return;

      /* retrieve the user app frame number from the pending queue */
      frame = zr36057->pend[zr36057->dma_tail & MJPEG_MASK_FRAME]; 
      /* get a pointer there to make accessing easier */
      gbuf = &zr36057->gbuf[frame];
      /* set a nice little timestamp on it */
      get_fast_time(&gbuf->bs.timestamp);

      if (zr36060->state == ZR_COMPRESS_STATE)
      {
         gbuf->bs.length = (stat_com & 0x7fffff) >> 1;

         /* update sequence number with the help of the counter in stat_com */
         seq = stat_com >> 24; /* the result stat_com has been marked by the ZR36057 */ 
         dif = (seq - zr36057->seq_num) & 0xff; /* little calculations to get the "real" seq_num */
         zr36057->seq_num += dif;
      }
      else
      {
	gbuf->bs.length = 0; /* decompress: show that this buffer has been serviced */
      }
      /* In the odd/even field mode the ZR36057 counts "wrong" */
      gbuf->bs.seq = zr36060->params.TmpDcm==2 ? (zr36057->seq_num>>1) : zr36057->seq_num;
      gbuf->state  = MJPEG_STATE_DONE; /* this buffer has been serviced */

      zr36057->dma_tail++; /* advance the dma queue */
   }
}









