
/*
 * INPUT FROM DISK (using RAW-IO if possible)
 *
 * Copyright (C) 1999  Thomas Mirlacher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author may be reached as dent@cosy.sbg.ac.at, or
 * Thomas Mirlacher, Jakob-Haringerstr. 2, A-5020 Salzburg,
 * Austria
 *
 *------------------------------------------------------------
 *
 * 15 Apr 2000
 *	Mirko Streckenbach <strecken@infosun.fmi.uni-passau.de>
 *	patch for RH6.X raw device support
 */


#define DVD_VIDEO_LB_LEN 2048

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#if defined (__NetBSD__) || defined(__OpenBSD__)
#include <sys/dvdio.h>
#elif defined(__linux__)
#include <linux/cdrom.h>
#else
#error "need to have DVD ioctls"
#endif

#include <oms/plugin.h>
#include <oms/oms_fileio.h>
#include "css/css.h"
#include "linux_raw.h"

static int udf_open (char *name);
static int udf_close (void);
static int udf_read (buf_t *buf, buf_entry_t *buf_entry);
static int udf_seek (u_long pos);
static int udf_open_css (void);
static int udf_read_css (buf_t *buf, buf_entry_t *buf_entry);

static struct {
	oms_off_t pos;
	int fd_raw;
	int fd;

	char key_disk[DVD_VIDEO_LB_LEN];
	char key_title[5];
} priv;

static plugin_input_t input = {
	udf_open,		// open
	udf_close,		// close
	udf_read,		// read
	udf_seek,		// seek
	DVD_VIDEO_LB_LEN,	// blocksize
	&priv			// private data
};


/**
 *
 */

static int udf_open_css (void)
{
	if (CSSAuthTitle (priv.fd, priv.key_title, priv.pos) < 0)
		perror ("Authenticate title");

	if (CSSDecryptTitleKey (priv.key_title, priv.key_disk) < 0)
		perror ("Decrypting title");

	return 0;
}


/**
 * this function is responsible for setting up the device to read data from,
 * as well as authorizing the area of the disc ...
 * well, this sounds wired, we have to set the filepointer BEFORE we open
 *  	the device - probaly add a function to request access to a specific 
 *	area, instead of having this hack, have to chekck what's more overhead
 *	1) have a somewhat strange open call, and beeing able to only open
 *		one device a time,
 *	2) add a call for requesting access, probably unused for most of the 
 *		functions
 */

static int udf_open (char *name)
{
	if ((priv.fd = open (name, O_RDONLY)) < 0) {
		perror ("opening udf disk");
		return -1;
	}

	if ((priv.fd_raw = raw_open (name)) < 0) {
		priv.fd_raw = priv.fd;
	}

// set filepointer to the right position, to make sure we request
//	authorization for the right area ...

	input.seek (priv.pos);

	if (is_encrypted ()) {
		if (CSSAuthDisk (priv.fd, priv.key_disk) < 0) {
			perror ("CSSAuthDisk");
			return -1;
		}	
				
		udf_open_css ();	
		input.read = udf_read_css;
	}

	return 0;
}


/**
 *
 */

static int udf_close (void)
{
	int i;
	dvd_authinfo ai;

	if (priv.fd != priv.fd_raw)
		close (priv.fd_raw);

// reset all AGIDs on the DVD drive
// dent: move this to the css subdir

	for (i=0; i<4; i++) {
		memset (&ai,  0, sizeof (dvd_authinfo));
		ai.type = DVD_INVALIDATE_AGID;
		ai.lsa.agid = i;
		ioctl (priv.fd, DVD_AUTH, &ai);
	}

	close (priv.fd);
	priv.fd = -1;

// if disk access has been authorized, reset again ...

	return 0;
}


/**
 *
 */

static int udf_read_css (buf_t *buf, buf_entry_t *buf_entry)
{
	int ret;

	priv.pos++;

        if ((ret = read (priv.fd_raw, buf_entry->mem, DVD_VIDEO_LB_LEN)) < 0)
		perror ("UDF");

// do the CSS stuff in SW, just for now
	CSSDescramble (buf_entry->mem, priv.key_title);

	return ret;
}


/**
 *
 */

static int udf_read (buf_t *buf, buf_entry_t *buf_entry)
{
	int ret;

	priv.pos++;

        if ((ret = read (priv.fd_raw, buf_entry->mem, DVD_VIDEO_LB_LEN)) < 0)
		perror ("UDF");

	return ret;
}


/**
 *
 */

static int udf_seek (u_long pos)
{
	priv.pos = pos;

	if (priv.fd >= 0) {
        	if (oms_lseek (priv.fd_raw, (oms_off_t) (pos)*DVD_VIDEO_LB_LEN, SEEK_SET) == -1) {
                	fprintf (stderr, "%s/%d: error in lseek at pos 0x%Lx\n",
                        	__FILE__, __LINE__, priv.pos);
                	return -1;
		}
        }

	return 0;
}


/**
 * move this out ... 
 */

static int is_encrypted (void)
{
	dvd_struct dvd;

	dvd.copyright.type = DVD_STRUCT_COPYRIGHT;
	dvd.copyright.layer_num = 0;

	if (ioctl (priv.fd, DVD_READ_STRUCT, &dvd) < 0) {
		perror ("DVD_READ_STRUCT");
		return -1;
	}

	return dvd.copyright.cpst;	// 0 ... not encrypted
}


/**
 * Initialize Plugin. 
 */

void *plugin_init (char *whoami)
{
	pluginRegister (whoami,
		PLUGIN_ID_INPUT,
		0,
		&input);

	return &input;
}


/**
 * Cleanup Plugin.
 */

void plugin_exit (void)
{
}
