/*
 * dvdudf: parse and read the UDF volume information of a DVD Video
 * Copyright (C) 1999 Christian Wolff for convergence integrated media GmbH
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 * 
 * The author can be reached at scarabaeus@convergence.de, 
 * the project's page is at http://linuxtv.org/dvd/
 */
 
#include "dvd_udf.h"

#include <stdio.h>
#include <string.h>

#ifndef u8
#define u8 unsigned char
#endif

#ifndef u16
#define u16 unsigned int
#endif

#ifndef u32
#define u32 unsigned long int
#endif

#ifndef u64
#define u64 unsigned long long int
#endif

#ifndef NULL
#define NULL ((void *)0)
#endif

#define MAX_FILE_LEN 2048

FILE* dvdromfile=NULL;  // CD/DVD-ROM block device or image file

struct Partition {
  int valid;
  char VolumeDesc[128];
  u16 Flags;
  u16 Number;
  char Contents[32];
  u32 AccessType;
  u32 Start;
  u32 Length;
} partition;

struct AD {
  u32 Location;
  u32 Length;
  u8 Flags;
  u16 Partition;
};

// for direct data access, LSB first
#define GETN1(p) ((u8)data[p])
#define GETN2(p) ((u16)data[p]|((u16)data[(p)+1]<<8))
#define GETN4(p) ((u32)data[p]|((u32)data[(p)+1]<<8)|((u32)data[(p)+2]<<16)|((u32)data[(p)+3]<<24))
#define GETN(p,n,target) memcpy(target,&data[p],n)

// reads absolute Logical Block of the disc
// returns number of read bytes on success, 0 on error
int UDFReadLB(unsigned long int lb_number, unsigned int block_count, unsigned char *data) {
	fpos64_t pos;

	if (dvdromfile==NULL)
		return 0;

	pos=(fpos64_t)lb_number*(fpos64_t)DVD_VIDEO_LB_LEN;

	if (fsetpos64(dvdromfile,&pos)<0)
		return 0; // position not found

	return fread(data,block_count*DVD_VIDEO_LB_LEN,1,dvdromfile);
}

int Unicodedecode(u8 *data, int len, char *target) {
	int p=1,i=0;

	if (!(data[0] & 0x18)) {
		target[0] ='\0';
		return 0;
	}

	if (data[0] & 0x10) {		// ignore MSB of unicode16
		p++;

		while (p<len) {
			target[i++]=data[p+=2];
		}
	} else {
		while (p<len) {
			target[i++]=data[p++];
		}
	}
	
	target[i]='\0';

	return 0;
}

int UDFEntity(u8 *data, u8 *Flags, char *Identifier) {
  Flags[0]=data[0];
  strncpy(Identifier,&data[1],5);
  return 0;
}

int UDFDescriptor(u8 *data, u16 *TagID) {
  TagID[0]=GETN2(0);
  // TODO: check CRC 'n stuff
  return 0;
}

int UDFExtentAD(u8 *data, u32 *Length, u32 *Location) {
  Length[0]  =GETN4(0);
  Location[0]=GETN4(4);
  return 0;
}

int UDFShortAD(u8 *data, struct AD *ad) {
  ad->Length=GETN4(0);
  ad->Flags=ad->Length>>30;
  ad->Length&=0x3FFFFFFF;
  ad->Location=GETN4(4);
  ad->Partition=partition.Number;  // use number of current partition
  return 0;
}

int UDFLongAD(u8 *data, struct AD *ad) {
  ad->Length=GETN4(0);
  ad->Flags=ad->Length>>30;
  ad->Length&=0x3FFFFFFF;
  ad->Location=GETN4(4);
  ad->Partition=GETN2(8);

  return 0;
}

int UDFExtAD(u8 *data, struct AD *ad) {
  ad->Length=GETN4(0);
  ad->Flags=ad->Length>>30;
  ad->Length&=0x3FFFFFFF;
  ad->Location=GETN4(12);
  ad->Partition=GETN2(16);

  return 0;
}

int UDFICB(u8 *data, u8 *FileType, u16 *Flags) {
  FileType[0]=GETN1(11);
  Flags[0]=GETN2(18);

  return 0;
}


int UDFPartition(u8 *data, u16 *Flags, u16 *Number, char *Contents, u32 *Start, u32 *Length) {
  Flags[0]=GETN2(20);
  Number[0]=GETN2(22);
  GETN(24,32,Contents);
  Start[0]=GETN4(188);
  Length[0]=GETN4(192);

  return 0;
}

// reads the volume descriptor and checks the parameters
// returns 0 on OK, 1 on error
int UDFLogVolume(u8 *data, char *VolumeDescriptor) {
  u32 lbsize,MT_L,N_PM;
  Unicodedecode(&data[84],128,VolumeDescriptor);
  lbsize=GETN4(212);  // should be 2048
  MT_L=GETN4(264);  // should be 6
  N_PM=GETN4(268);  // should be 1

  if (lbsize!=DVD_VIDEO_LB_LEN)
	return 1;

  return 0;
}

int UDFFileEntry(u8 *data, u8 *FileType, struct AD *ad) {
  u8 filetype;
  u16 flags;
  u32 L_EA,L_AD;
  int p;

  UDFICB(&data[16],&filetype,&flags);
  FileType[0]=filetype;
  L_EA=GETN4(168);
  L_AD=GETN4(172);
  p=176+L_EA;
  while (p<176+L_EA+L_AD) {
    switch (flags&0x0007) {
      case 0: UDFShortAD(&data[p],ad); p+=8;  break;
      case 1: UDFLongAD(&data[p],ad);  p+=16; break;
      case 2: UDFExtAD(&data[p],ad);   p+=20; break;
      case 3:
        switch (L_AD) {
          case 8:  UDFShortAD(&data[p],ad); break;
          case 16: UDFLongAD(&data[p],ad);  break;
          case 20: UDFExtAD(&data[p],ad);   break;
        }
        p+=L_AD;
        break;
      default: p+=L_AD; break;
    }
  }
  return 0;
}

int UDFFileIdentifier(u8 *data, u8 *FileCharacteristics, char *FileName, struct AD *FileICB) {
  u8 L_FI;
  u16 L_IU;
  
  FileCharacteristics[0]=GETN1(18);
  L_FI=GETN1(19);
  UDFLongAD(&data[20],FileICB);
  L_IU=GETN2(36);
  if (L_FI) Unicodedecode(&data[38+L_IU],L_FI,FileName);
  else FileName[0]='\0';
  return 4*((38+L_FI+L_IU+3)/4);
}

// Maps ICB to FileAD
// ICB: Location of ICB of directory to scan
// FileType: Type of the file
// File: Location of file the ICB is pointing to
// return 1 on success, 0 on error;
int UDFMapICB(struct AD ICB, u8 *FileType, struct AD *File) {
  u8 LogBlock[DVD_VIDEO_LB_LEN];
  u32 lbnum;
  u16 TagID;

  lbnum=partition.Start+ICB.Location;
  do {
    if (!UDFReadLB(lbnum++,1,LogBlock)) TagID=0;
    else UDFDescriptor(LogBlock,&TagID);
    if (TagID==261) {
      UDFFileEntry(LogBlock,FileType,File);
      return 1;
    };
  } while ((lbnum<=partition.Start+ICB.Location+(ICB.Length-1)/DVD_VIDEO_LB_LEN) && (TagID!=261));
  return 0;
}
  
// Dir: Location of directory to scan
// FileName: Name of file to look for
// FileICB: Location of ICB of the found file
// return 1 on success, 0 on error;
int UDFScanDir(struct AD Dir, char *FileName, struct AD *FileICB) {
  u8 LogBlock[DVD_VIDEO_LB_LEN];
  u32 lbnum;
  u16 TagID;
  u8 filechar;
  char filename[MAX_FILE_LEN];
  int p;
  
  // Scan dir for ICB of file
  lbnum=partition.Start+Dir.Location;
  do {
    if (!UDFReadLB(lbnum++,1,LogBlock)) TagID=0;
    else {
      p=0;
      while (p<DVD_VIDEO_LB_LEN) {
        UDFDescriptor(&LogBlock[p],&TagID);
        if (TagID==257) {
          p+=UDFFileIdentifier(&LogBlock[p],&filechar,filename,FileICB);
          if (!strcasecmp(FileName,filename)) return 1;
        } else p=DVD_VIDEO_LB_LEN;
      }
    }
  } while (lbnum<=partition.Start+Dir.Location+(Dir.Length-1)/DVD_VIDEO_LB_LEN);
  return 0;
}

// looks for partition on the disc
//   partnum: number of the partition, starting at 0
//   part: structure to fill with the partition information
//   return 1 if partition found, 0 on error;
int UDFFindPartition(int partnum, struct Partition *part) {
  u8 LogBlock[DVD_VIDEO_LB_LEN],Anchor[DVD_VIDEO_LB_LEN];
  u32 lbnum,MVDS_location,MVDS_length;
  u16 TagID;
  u32 lastsector;
  int i,terminate,volvalid;

#if 0
/************************/
  lbnum=0x20;

  UDFReadLB(lbnum,1,LogBlock);

{int i;

fprintf (stderr, "ID: ");
//for (i=0; i<128; i++)
	fprintf (stderr, "%s", &LogBlock[i+72]);
fprintf (stderr, "\n");
}

/************************/
#endif

  // Find Anchor
  lastsector=0;
  lbnum=256;   // try #1, prime anchor
  terminate=0;
  while (1) {  // loop da loop
    if (UDFReadLB(lbnum,1,Anchor)) {
      UDFDescriptor(Anchor,&TagID);
    } else TagID=0;
    if (TagID!=2) {             // not an anchor?
      if (terminate) return 0;  // final try failed 
      if (lastsector) {         // we already found the last sector
        lbnum=lastsector;       // try #3, alternative backup anchor
        terminate=1;            // but that's just about enough, then!
      } else {
        // TODO: find last sector of the disc (this is optional)
        if (lastsector) lbnum=lastsector-256; // try #2, backup anchor
        else return 0;          // unable to find last sector
      }
    } else break;               // it is an anchor! continue...
  }
  UDFExtentAD(&Anchor[16],&MVDS_length,&MVDS_location);  // main volume descriptor
  
  part->valid=0;
  volvalid=0;
  part->VolumeDesc[0]='\0';
  i=1;
  do {
    // Find Volume Descriptor
    lbnum=MVDS_location;
    do {
      if (!UDFReadLB(lbnum++,1,LogBlock)) TagID=0;
      else UDFDescriptor(LogBlock,&TagID);
      if ((TagID==5) && (!part->valid)) {  // Partition Descriptor
        UDFPartition(LogBlock,&part->Flags,&part->Number,part->Contents,&part->Start,&part->Length);
        part->valid=(partnum==part->Number);
      } else if ((TagID==6) && (!volvalid)) {  // Logical Volume Descriptor
        if (UDFLogVolume(LogBlock,part->VolumeDesc)) {  
          //TODO: sector size wrong!
        } else volvalid=1;
      }
    } while ((lbnum<=MVDS_location+(MVDS_length-1)/DVD_VIDEO_LB_LEN) && (TagID!=8) && ((!part->valid) || (!volvalid)));
    if ((!part->valid) || (!volvalid)) UDFExtentAD(&Anchor[24],&MVDS_length,&MVDS_location);  // backup volume descriptor
  } while (i-- && ((!part->valid) || (!volvalid)));
  return (part->valid);  // we only care for the partition, not the volume
}

// looks for a file on the UDF disc/imagefile
// filename has to be the absolute pathname on the UDF filesystem, starting with /
// returns absolute LB number, or 0 on error
unsigned long int UDFFindFile(char *filename) {
  u8 LogBlock[DVD_VIDEO_LB_LEN];
  u32 lbnum;
  u16 TagID;
  struct AD RootICB,File,ICB;
  char tokenline[MAX_FILE_LEN];
  char *token;
  u8 filetype;
  
  int Partition=0;  // this is the standard location for DVD Video
  
  tokenline[0]='\0';
  // Very DVD-Video specific:
  //if (filename[0]!='/') strcat(tokenline,"/VIDEO_TS/");
  strcat(tokenline,filename);

  // Find partition
  if (!UDFFindPartition(Partition,&partition)) return 0;
  
  // Find root dir ICB
  lbnum=partition.Start;
  do {
    if (!UDFReadLB(lbnum++,1,LogBlock)) TagID=0;
    else UDFDescriptor(LogBlock,&TagID);
    if (TagID==256) {  // File Set Descriptor
      UDFLongAD(&LogBlock[400],&RootICB);
    }
  } while ((lbnum<partition.Start+partition.Length) && (TagID!=8) && (TagID!=256));
  if (TagID!=256) return 0;
  if (RootICB.Partition!=Partition) return 0;
  
  // Find root dir
  if (!UDFMapICB(RootICB,&filetype,&File)) return 0;
  if (filetype!=4) return 0;  // root dir should be dir

  // Tokenize filepath
  token=strtok(tokenline,"/");
  while (token != NULL) {
    if (!UDFScanDir(File,token,&ICB)) return 0;
    if (!UDFMapICB(ICB,&filetype,&File)) return 0;
    token=strtok(NULL,"/");
  }
  return partition.Start+File.Location;
}

// open block device or image file
// returns fileno() of the file on success, or -1 on error
int UDFOpenDisc(char *filename) {
  if ((dvdromfile=fopen(filename,"r"))==NULL) return -1;
  return fileno(dvdromfile);
}
